#define _GNU_SOURCE
#include <sched.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <sys/eventfd.h>
#include <sys/mman.h>

static volatile bool terminated;

typedef struct
{
    pthread_t thread;
    int id;
    int trigger_fd;
}
thread_data_t;

void *thread_func(void *arg)
{
    int rc;
    thread_data_t *thr = arg;
    int loop;
    {
        struct sched_param param;
        memset(&param, 0, sizeof(param));
        param.sched_priority = 60;
        rc = pthread_setschedparam(pthread_self(), SCHED_FIFO, &param);
        if (rc) {
            fprintf(stderr, "pthread_setschedparam %m\n");
            exit(1);
        }

        /* Run each thread on a separate CPU core */
        cpu_set_t set;
        CPU_ZERO(&set);
        CPU_SET(thr->id, &set);
        rc = sched_setaffinity(0, sizeof(set), &set);
        if (rc) {
            fprintf(stderr, "sched_setaffinity %m\n");
            exit(1);
        }
    }

    for (loop=0; !terminated ; loop++) {
        uint64_t u64;
        int rc = read(thr->trigger_fd, &u64, 8);
        if (rc > 0) {
            /* printf("thread %i loop %i\n", thr->id, loop); */
            system("lspci -xxx > /dev/null");
        }
        else {
            fprintf(stderr, "read %m\n");
        }
    }
    printf("End of thread %i\n", thr->id);
    return NULL;
}

#define NO_THREADS 4
static thread_data_t thread_data[NO_THREADS];

int main(int argc, char *argv[])
{
    int i;
    int rc;
    /* Init */
    mlockall(MCL_CURRENT | MCL_FUTURE);
    for (i=0; i<NO_THREADS; i++) {
        thread_data[i].id = i;
        rc = thread_data[i].trigger_fd = eventfd(0,0);
        if (rc < 0) {
            fprintf(stderr, "Error creating eventfd: %m\n"); 
            return 1;
        }
    }

    for (i=0; i<NO_THREADS; i++) {
        rc = pthread_create(&thread_data[i].thread, NULL, thread_func, &thread_data[i]);
        if (rc) {
            fprintf(stderr, "Error creating thread: %m\n");
            return 1;
        }
    }

    /* Loop */
    int j;
    int max = 100;
    for (j=0; j<=max; j++) {
        struct timespec ts = { 0, 10000000 };
        clock_nanosleep(CLOCK_MONOTONIC, 0, &ts, NULL);

        if (j == max) {
            terminated = true;
        }

        for (i=0; i<NO_THREADS; i++) {
            uint64_t u64 = 1;
            write(thread_data[i].trigger_fd, &u64, 8);
        }
    }
    
    /* Cleanup */
    for (i=0; i<NO_THREADS; i++) {
        pthread_join(thread_data[i].thread, NULL);
        printf("Thread %i joined\n", i);
        close(thread_data[i].trigger_fd);
    }
    return 0;
}

